import time
from typing import Callable, Final, Literal, Optional, TypeVar
from os import urandom
from numba import njit
from numpy import uint64, int64

try:
    from . import omnim_rng_rust

    HAS_RUST = True
except ImportError:
    HAS_RUST = False

T = TypeVar("T")


def has_rust():
    return HAS_RUST


@njit
def _xorshift64_next(state: int) -> int:
    x = state
    x ^= (x << 13) & 0xFFFFFFFFFFFFFFFF
    x ^= (x >> 7) & 0xFFFFFFFFFFFFFFFF
    x ^= (x << 17) & 0xFFFFFFFFFFFFFFFF
    return x


@njit
def _pcg32_next(state: int, inc: int):
    oldstate = uint64(state)
    new_state = oldstate * uint64(6364136223846793005) + uint64(inc)

    shifted = uint64(((oldstate >> 18) ^ oldstate) >> 27)
    rot = oldstate >> 59
    output = (shifted >> rot) | (shifted << (uint64(-rot) & uint64(31)))

    return int64(new_state), int64(output & uint64(0xFFFFFFFF))


class _xorshift64:
    def __init__(self, seed: int):
        if HAS_RUST:
            self.impl = omnim_rng_rust.Xorshift64(0)
        else:
            self.state: int = seed & 0xFFFFFFFFFFFFFFFF or 0xDEADBEEF
        self.max: Final = 0xFFFFFFFFFFFFFFFF

    def next(self) -> int:
        if HAS_RUST:
            self.state = self.impl.next()
        else:
            self.state = _xorshift64_next(self.state)
        return self.state

    def nexts(self, count: int, min_value: int, max_value: int) -> list[int]:
        self.state = self.impl.next_ints(count, min_value, max_value)
        return self.state


class _pcg32:
    def __init__(self, seed: int):
        if HAS_RUST:
            self.impl = omnim_rng_rust.Pcg32(0)
        else:
            self.state = (seed + 0xDA3E39CB94B95BDB) & 0xFFFFFFFFFFFFFFFF
            self.inc = (seed | 1) & 0xFFFFFFFFFFFFFFFF
        self.max: Final = 0xFFFFFFFF

    def next(self) -> int:
        if HAS_RUST:
            self.state = result = self.impl.next()
        else:
            self.state, result = _pcg32_next(self.state, self.inc)
        return result

    def nexts(self, count: int, min_value: int, max_value: int) -> list[int]:
        self.state = self.impl.next_ints(count, min_value, max_value)
        return self.state


class rng:
    def __init__(
        self,
        *,
        seed: Optional[int] = None,
        mode: Literal["xorshift", "pcg"] = "xorshift",
    ):
        import time

        if seed is None:
            seed = int(time.time() * 1000)

        self.__generator = None
        match mode:
            case "xorshift":
                self.__generator = _xorshift64(seed)
            case "pcg":
                self.__generator = _pcg32(seed)

    def _next(self) -> int:
        return self.__generator.next()

    def next_int(self, min_value: int, max_value: int) -> int:
        return self._next() % (max_value - min_value + 1) + min_value

    def next_float(self, min_value: float, max_value: float) -> float:
        return self._next() / self.__generator.max * (max_value - min_value) + min_value

    def next_ints(self, count: int, min_value: int, max_value: int) -> list[int]:
        if HAS_RUST:
            raw_values = self.__generator.impl.next_batch(count)
            range_size = max_value - min_value + 1
            return [val % range_size + min_value for val in raw_values]
            # return self.__generator.impl.next_ints(count, min_value, max_value)
        else:
            range_size = max_value - min_value + 1
            return [(self._next() % range_size + min_value) for _ in range(count)]

    def next_floats(
        self, count: int, min_value: float, max_value: float
    ) -> list[float]:
        if HAS_RUST:
            raw_values = self.__generator.impl.next_batch(count)
            max_val = self.__generator.max
            diff = max_value - min_value
            return [(val / max_val * diff + min_value) for val in raw_values]
        else:
            max_val = self.__generator.max
            diff = max_value - min_value
            return [(self._next() / max_val * diff + min_value) for _ in range(count)]

    def rnexts(self, count: int, min_value: int, max_value: int) -> list[int]:
        return self.__generator.impl.next_ints(count, min_value, max_value)

    def runiforms(self, count: int, min_value: float, max_value: float) -> list[float]:
        return self.__generator.impl.next_floats(count, min_value, max_value)

    def choice(self, seq: list[T]):
        if not seq:
            raise IndexError("Cannot choose from an empty sequence")
        return seq[self.next_int(0, len(seq) - 1)]


try:
    _g = rng(seed=int.from_bytes(urandom(8), "big"), mode="pcg")
except Exception:  # NotImplementedError
    _g = rng()

T = TypeVar("T", int, float)

uniform = _g.next_float
uniforms = _g.next_floats
randint = _g.next_int
randints = _g.next_ints
rnexts = _g.rnexts
runiforms = _g.runiforms


def rand(min: T, max: T) -> T:
    if isinstance(min, int) and isinstance(max, int):
        return _g.next_int(min, max)
    return _g.next_float(min, max)


if __name__ == "__main__":
    _rng = rng(seed=10)
    print(_rng.next_int(0, 100))
    print(_rng.next_float(0.0, 1.0))
    print(_rng.choice([1, 2, 3, 4, 5]))

    print([rand(0, 3) for _ in range(10)])
    print([rand(0, 1.0) for _ in range(10)])



from . import delegate
from . import rx
from . import nullable
from . import readonly
from . import linq
from . import task

__all__ = [
    "delegate",
    "rx",
    "nullable",
    "readonly",
    "linq",
    "task",
]



import asyncio
from typing import Awaitable, Callable, Optional, overload

global __interval_s
__interval_s = 0.1


@overload
def interval(value: float) -> None:
    """value is in seconds (e.g., 0.1)"""
    ...


@overload
def interval(value: int) -> None:
    """value is in milliseconds (e.g., 100)"""
    ...


def interval(value: float | int) -> None:
    global __interval_s
    match value:
        case float() as s:
            __interval_s = max(0.001, s)
        case int() as ms:
            __interval_s = max(0.001, ms / 1000.0)


async def wait_until(
    predicate: Callable[[], bool],
    timeout: Optional[float] = None,
) -> bool:
    loop = asyncio.get_running_loop()
    start = loop.time()

    while not predicate():
        if timeout is not None:
            if (loop.time() - start) > timeout:
                raise TimeoutError()

        await asyncio.sleep(__interval_s)

    return True


async def wait_while(
    predicate: Callable[[], bool],
    timeout: Optional[float] = None,
) -> bool:
    return await wait_until(lambda: not predicate(), timeout)


async def wait_all(*tasks: Awaitable):
    return await asyncio.gather(*tasks)


async def wait_any(*tasks: Awaitable):
    done, pending = await asyncio.wait(
        [asyncio.ensure_future(t) for t in tasks],
        return_when=asyncio.FIRST_COMPLETED,
    )
    return done, pending


async def with_timeout(task: Awaitable, timeout: float):
    return await asyncio.wait_for(task, timeout)


async def wait_frame():
    await asyncio.sleep(0)


async def delay(seconds: float):
    await asyncio.sleep(seconds)



from typing import (
    Any,
    Callable,
    Generator,
    Generic,
    Iterable,
    Iterator,
    Optional,
    TypeVar,
)
from builtins import range as _range
from builtins import any as _any
from builtins import zip as _zip

T_SOURCE = TypeVar("T_SOURCE")
T_RESULT = TypeVar("T_RESULT")

T_ZIP_INNER = TypeVar("T_ZIP_INNER")
T_ZIP_OUTER = TypeVar("T_ZIP_OUTER")
T_ZIP_RESULT = TypeVar("T_ZIP_RESULT")

T_JOIN_INNER = TypeVar("T_JOIN_INNER")
T_JOIN_SELECTOR_RESULT = TypeVar("T_JOIN_SELECTOR_RESULT")
T_JOIN_RESULT = TypeVar("T_JOIN_RESULT")


class linq(Generic[T_SOURCE]):
    def __init__(self, src: Iterator[T_SOURCE] | Iterable[T_SOURCE]) -> None:
        self.__src = src

    def __iter__(self) -> Iterator[T_SOURCE]:
        return iter(self.__src)

    def count(self, pred: Optional[Callable[[T_SOURCE], bool]] = None) -> int:
        return count(self.__src, pred)

    def any(self, pred: Optional[Callable[[T_SOURCE], bool]] = None) -> bool:
        return any(self, pred)

    def where(self, pred: Callable[[T_SOURCE], bool]) -> "linq[T_SOURCE]":
        return linq(where(self, pred))

    def select(self, pred: Callable[[T_SOURCE], T_RESULT]) -> "linq[T_RESULT]":
        return linq(select(self, pred))

    def zip(
        self,
        other: Iterable[T_ZIP_OUTER],
        selector: Callable[[T_SOURCE, T_ZIP_OUTER], T_ZIP_RESULT],
    ) -> "linq[T_ZIP_RESULT]":
        return linq(zip(self.__src, other, selector))

    def join(
        self,
        inner: Iterable[T_JOIN_INNER],
        outer_selector: Callable[[T_SOURCE], T_JOIN_SELECTOR_RESULT],
        inner_selector: Callable[[T_JOIN_INNER], T_JOIN_SELECTOR_RESULT],
        result_selector: Callable[[T_SOURCE, T_JOIN_INNER], T_JOIN_RESULT],
    ) -> "linq[T_JOIN_RESULT]":
        return linq(join(self, inner, outer_selector, inner_selector, result_selector))

    def first(
        self,
        pred: Optional[Callable[[T_SOURCE], bool]] = None,
    ) -> Optional[T_SOURCE]:
        return first(self, pred)

    def last(self, pred: Optional[Callable[[T_SOURCE], bool]] = None) -> T_SOURCE:
        return last(self, pred)

    def order_by(
        self,
        selector: Callable[[T_SOURCE], Any],
        descending: bool = False,
    ) -> "linq[T_SOURCE]":
        return linq(iter(order_by(self, selector, descending)))

    def as_list(self) -> list[T_SOURCE]:
        return as_list(self)


def count(
    src: Iterable[T_SOURCE],
    pred: Optional[Callable[[T_SOURCE], bool]] = None,
) -> int:
    if pred:
        return sum(1 for e in src if pred(e))
    else:
        return sum(1 for _ in src)


def any(
    src: Iterable[T_SOURCE],
    pred: Optional[Callable[[T_SOURCE], bool]] = None,
) -> bool:
    for e in src:
        if (not pred) or (pred and pred(e)):
            return True
    return False


def where(
    src: Iterable[T_SOURCE],
    pred: Callable[[T_SOURCE], bool],
) -> Iterable[T_SOURCE]:
    return iter(e for e in src if pred(e))


def select(
    src: Iterable[T_SOURCE],
    pred: Callable[[T_SOURCE], T_RESULT],
) -> Iterable[T_RESULT]:
    return iter(pred(e) for e in src)


def zip(
    src: Iterable[T_SOURCE],
    other: Iterable[T_ZIP_OUTER],
    selector: Callable[[T_SOURCE, T_ZIP_OUTER], T_ZIP_RESULT],
) -> Iterable[T_ZIP_RESULT]:
    return iter(selector(x, y) for x, y in _zip(src, other))


def join(
    src: Iterable[T_SOURCE],
    inner: Iterable[T_JOIN_INNER],
    outer_selector: Callable[[T_SOURCE], T_JOIN_SELECTOR_RESULT],
    inner_selector: Callable[[T_JOIN_INNER], T_JOIN_SELECTOR_RESULT],
    result_selector: Callable[[T_SOURCE, T_JOIN_INNER], T_JOIN_RESULT],
) -> Iterable[T_JOIN_RESULT]:
    def generator():
        inner_table = {}
        for item in inner:
            inner_table.setdefault(inner_selector(item), []).append(item)

        for oi in src:
            if (ok := outer_selector(oi)) in inner_table:
                for ii in inner_table[ok]:
                    yield result_selector(oi, ii)

    return generator()


def first(
    src: Iterable[T_SOURCE],
    pred: Optional[Callable[[T_SOURCE], bool]] = None,
) -> Optional[T_SOURCE]:
    for e in src:
        if pred is None or pred(e):
            return e
    return None


def last(self, pred: Optional[Callable[[T_SOURCE], bool]] = None) -> T_SOURCE:
    if pred is None:
        try:
            return next(reversed(self.__src))  # type: ignore
        except (TypeError, StopIteration) as e:
            raise e

    last_item = None
    found = False
    for e in self:
        if pred is None or pred(e):
            last_item = e
            found = True

    if not found:
        raise ValueError()
    return last_item  # type: ignore


def order_by(
    src: Iterable[T_SOURCE],
    selector: Callable[[T_SOURCE], Any],
    descending: bool = False,
) -> Iterable[T_SOURCE]:
    def generator():
        items = list(src)
        items.sort(key=selector, reverse=descending)
        for item in items:
            yield item

    return generator()


def as_linq(src: Iterable[T_SOURCE]) -> linq[T_SOURCE]:
    return linq(src)


def as_list(src: Iterable[T_SOURCE]) -> list[T_SOURCE]:
    return list(src)


def repeat(e: T_SOURCE, count: int) -> linq[T_SOURCE]:
    return linq(e for _ in _range(count))


def range(start: int, count: int) -> linq[int]:
    return linq(i for i in _range(start, start + count))



import builtins
from math import floor as _floor
import sys
from typing import Final, TypeVar

_builtins_min = builtins.min
_builtins_max = builtins.max

T_NUM = TypeVar("T_NUM", int, float)

pi: Final = 3.14159265358979323846
e: Final = 2.718281828459045235360287471352
eps: Final = 1 - (((4 / 3) - 1) + ((4 / 3) - 1) + ((4 / 3) - 1))

rad_to_deg: Final = 180 / pi
deg_to_rad: Final = pi / 180


def to_degrees(x: float) -> float:
    return x * rad_to_deg


def to_radians(x: float) -> float:
    return x * deg_to_rad


def sqrt(x: T_NUM) -> float:
    return x ** (1 / 2)


def min(*xs: T_NUM) -> T_NUM:
    return _builtins_min(*xs)


def max(*xs: T_NUM) -> T_NUM:
    return _builtins_max(*xs)


def clamp(x: T_NUM, min_value: T_NUM, max_value: T_NUM) -> T_NUM:
    if x < min_value:
        return min_value
    elif x > max_value:
        return max_value
    return x


def sign(x: T_NUM) -> int:
    if x == 0:
        return 0
    elif x > 0:
        return 1
    else:
        return -1


def abs(a: T_NUM) -> T_NUM:
    return -a if a < 0 else a


def floor(x: T_NUM, d=0) -> T_NUM:
    def _inner_floor(x: T_NUM) -> T_NUM:
        return type(x)(x - (x % 1))

    if d != 0:
        p = 10**d
        return type(x)(_inner_floor(x * p) / p)

    return _inner_floor(x)


def round(x: T_NUM, d: int = 0) -> T_NUM:
    p = 10**d
    res = _floor(x * p + 0.5) / p
    return type(x)(res)


def is_prime(n: int) -> bool:
    if n == 2:
        return True
    elif n < 2 or n % 2 == 0:
        return False

    for i in range(3, _floor(sqrt(n)) + 1, 2):
        if n % i == 0:
            return False
    return True



from typing import Final, Generic, Iterable, Iterator, TypeAlias, TypeVar
from .math import abs

_Index: TypeAlias = int | float  # TypeVar("_Index", int, float)
T_INDEX = TypeVar("T_INDEX", int, float)


class frange(Generic[T_INDEX]):
    """
    flexible range

    ```python
    frange(0, 2, 1) # [0, 1, 2]
    frange(0, 2, 1.0) # [0.0, 1.0, 2.0]
    frange(0, 2.0, 1) # [0.0, 1.0, 2.0]
    frange(0.0, 2.0, 1) # [0.0, 1.0, 2.0]
    ```
    """

    def __init__(self, begin: T_INDEX, end: T_INDEX, step: T_INDEX) -> None:
        self.__begin = begin
        self.__end = end
        self.__step = step
        self.__forward: Final = self.__begin < self.__end

    def __iter__(self):
        if (step := abs(self.__step)) == 0:
            raise ValueError("step must not be zero")

        value = self.__begin
        if self.__forward:
            while value <= self.__end:
                yield (value)
                value += step
        else:
            while value >= self.__end:
                yield (value)
                value -= step


T = TypeVar("T")


class step:
    """
    ## with iterator:
    ```python
    items = [0, 2, 4, 6, 8]
    for i in step(items):
        print(i) # [0, 1, 2, 3, 4]
    ```

    ## with count:
    ```python
    for i in step(3):
        print(i) # [0, 1, 2]
    ```
    """

    def __init__(self, src, *, reverse=False) -> None:
        if isinstance(src, Iterable):
            self.__len = len(list(iter(src)))
        elif isinstance(src, int):
            self.__len = src
        self.__reverse = reverse

    def __iter__(self) -> Iterator[int]:
        r = range(self.__len)
        if self.__reverse:
            r = reversed(r)
        for i in r:  # range(self.__len):
            yield i

    def __len__(self) -> int:
        return self.__len

    def __getitem__(self, i: int) -> int:
        return i


if __name__ == "__main__":
    print(f"-10, +10: {list(frange(-10, 10, -1))}")
    print(f"+10, -10: {list(frange(10, -10, -1))}")
    print(f"0.0, 1.0: {list(frange(0, 1, 0.1))}")
    _ = list(frange[float](0, 1, 0.1))

    print("---")

    src = [i for i in frange(0, 10, 1)]
    print(src)
    print(list(step(src)))

    print("---")

    items = [i for i in range(0, 10, 2)]
    for i in step(items):
        print(i)  # [0, 2, 4, 6, 8]



from typing import Optional


class _chunk:
    def __init__(self, cap=16) -> None:
        self.buffer: list[Optional[str]] = [None] * cap
        self.count = 0
        self.next: Optional[_chunk] = None


class stringbuilder:
    def __init__(self, chunk_cap=16) -> None:
        self.__chunk_cap = chunk_cap
        self.__head = _chunk(self.__chunk_cap)
        self.__tail = self.__head

    def append(self, arg: object) -> None:
        for c in str(arg):
            if self.__tail.count >= self.__chunk_cap:
                new_chunk = _chunk(self.__chunk_cap)
                self.__tail.next = new_chunk
                self.__tail = new_chunk
            self.__tail.buffer[self.__tail.count] = c
            self.__tail.count += 1

    def append_format(self, fmt: str, *args: object) -> None:
        self.append(fmt.format(*args))

    def append_join(self, sep: str, *args: object) -> None:
        self.append(sep.join(str(a) for a in args))

    def append_line(self, arg: object) -> None:
        last_char = ""
        if self.__tail.count > 0:
            last_char = self.__tail.buffer[-1]

        is_empty = self.__head.count == 0 and self.__head.next is None
        if not is_empty and last_char != "\n":
            self.append("\n")

        self.append(arg)

    def insert(self, idx: int, arg: object) -> None:
        if not arg:
            return

        current_chunk = self.__head
        remaining_idx = idx
        while current_chunk:
            if remaining_idx <= current_chunk.count:
                break
            remaining_idx -= current_chunk.count
            current_chunk = current_chunk.next
        else:
            self.append(arg)
            return

        tail_buffer = []
        temp_chunk = current_chunk
        tail_buffer.extend(temp_chunk.buffer[remaining_idx : temp_chunk.count])
        temp_chunk = temp_chunk.next
        while temp_chunk:
            tail_buffer.extend(temp_chunk.buffer[: temp_chunk.count])
            temp_chunk = temp_chunk.next

        current_chunk.count = remaining_idx
        current_chunk.next = None
        self.__tail = current_chunk

        self.append(arg)
        for item in tail_buffer:
            if item is not None:
                self.append(item)

    def remove(self, idx: int, length: int) -> None:
        if length <= 0:
            return

        current_str = str(self)
        if idx < 0 or idx > len(current_str):
            return

        new_str = current_str[:idx] + current_str[idx + length :]

        self.clear()
        self.append(new_str)

    def clear(self) -> None:
        self.__head = _chunk(self.__chunk_cap)
        self.__tail = self.__head

    def __str__(self) -> str:
        result = []
        cur = self.__head
        while cur:
            result.extend(cur.buffer[: cur.count])
            cur = cur.next
        return "".join(char for char in result if char is not None)

    def __repr__(self) -> str:
        length = len(text := str(self))
        preview = text if length <= 50 else text[:47] + "..."
        return f"<{self.__class__.__name__} length={length} value={repr(preview)}>"



from inspect import isfunction, isclass, ismethod, currentframe


def nameof(obj) -> str:
    if isfunction(obj) or isclass(obj) or ismethod(obj):
        return obj.__name__
    elif hasattr(obj, "__call__") and hasattr(obj, "__name__"):
        return obj.__name__
    else:
        frame = currentframe()
        if frame is None:
            return ""
        try:
            caller_frame = frame.f_back
            if caller_frame is None:
                return ""

            for name, value in caller_frame.f_locals.items():
                if value is obj:
                    return name
            for name, value in caller_frame.f_globals.items():
                if value is obj:
                    return name
        finally:
            del frame
        return ""


if __name__ == "__main__":

    def add(a, b) -> int:
        return a + b

    a = 2
    b = -3
    print(f"{nameof(add)}({nameof(a)}, {nameof(b)}): {add(a, b)}")
